/*
	xpreso

	Description: simple spread-sheet AP

	Operate: click button or press ESCAPE key to exit.

*/

#include <windows.h>
#include <winuser.h>
#include <wintools.h>
#include <device.h>
#include <xime.h>	// xime header
#include <resources.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <values.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>

#include "xpreso.h"
#include "x_bclib.h"

#define ROW_MAX 99
#define COL_MAX 26

typedef struct cell{
  int id;
  char data[64];              //raw data
  double value;               //processed value for numbers
  char disp[32];              //to display, max 6 char for numbers
  struct cell *next;
  struct cell *prev;
}cell;

cell* m_book[32][100];
cell* first_cell=NULL;
cell* last_cell=NULL;
cell m_cells[32];

int m_row=0, m_col=0;   //col # and row # for left , top 
int n_rows=13;
int n_cols=4;
int n_chars=6;
int m_x=0, m_y=0;  //curosr cell x,y within the window
int m_old_x=0, m_old_y=0;  //curosr cell x,y within the window
int m_entries=0;   //# of entries
int m_col_max=0, m_row_max=0;

FILE* data_file;
char filename[128]="";
char fullname[128]="";
char format_str[32]="";
char caption[64]="";
char m_range[8]="";
char m_key[4]="";

PALETTEENTRY pals[16];

extern NAMEINFO menu_info[];
extern NAMEINFO dialog_info[];
extern int SetScreenMode (int bpp);
extern void ResetScreen (void);
extern void showhelp( char name[], char *file );
extern int showabout( char * file );

HFONT GetSystemFontObject( void );
HMENU LoadMenuEx(NAMEINFO *minfo, LPCSTR lpMenuName,int count);
HWND DialogBoxEx (HINSTANCE, LPCSTR, HWND, DLGPROC, NAMEINFO *, int);
BOOL WriteClipboard (char *buf);
BOOL ReadClipboard (char *buf, int buf_size);
//long GetClipboardSize (void);
BOOL IsClipboardEmpty (void);
int RegisterAboutClass(void);
int WINAPI RegisterHelpClass (HINSTANCE hInstance);


LRESULT CALLBACK wproc (HWND, UINT, WPARAM, LPARAM);
static LRESULT CALLBACK LoadProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK SaveProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK SortProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK ChartProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK PieProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK LineProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK BarProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK JumpProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);

void MyPaint(HWND hwnd);
void enter_data(HWND hwnd);
void enter_data1(char *data);
void replace_cell(char* buf);
void del_cell(HWND hwnd);
void new_file();
void fix_values();
void fix_disp();
void load_file(FILE* file);
void save_file(FILE* file);
int is_text(char str[], int m_begin, int m_end);

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, 
		    PSTR szCmdLine, int iCmdShow){

  static char szAppName[] = "XPRESO - New File";
  
  HWND hwnd, hedit;
  MSG msg;
  WNDCLASS wndclass;
  
  wndclass.style	= CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
  wndclass.lpfnWndProc	= (WNDPROC) wproc;
  wndclass.cbClsExtra	= 0;
  wndclass.cbWndExtra	= 0;
  wndclass.hInstance	= 0;
  wndclass.hIcon	= 0;
  wndclass.hCursor	= 0;
  wndclass.hbrBackground = (HBRUSH) GetStockObject (LTGRAY_BRUSH);
  wndclass.lpszMenuName	= NULL;
  wndclass.lpszClassName = szAppName;

  RegisterClass (&wndclass);
  
  RegisterAboutClass();
  RegisterHelpClass( hInstance );

  memset(m_book, 0, sizeof(cell*)*3200);

  //main window
  hwnd = CreateWindowEx(0L,szAppName,"XPRESO - New File",
			WS_VISIBLE | WS_CAPTION,
			0 , 0, 160, 240,
			NULL, NULL, NULL, NULL);

  sprintf(caption, "XPRESO - New File");

  hwnd->hMenu   =LoadMenuEx(menu_info,"BRMENU",1);


  //input edit
  hedit = CreateWindowEx(0L,"EDIT","",
			 WS_CHILD | WS_VISIBLE | WS_BORDER,
			 55, 18, 105, 12,
			 hwnd, (HMENU) ID_EDIT, NULL, NULL);
  
  //exit button
  CreateWindowEx(0L, "BUTTON", "Quit",
		 WS_CHILD | WS_VISIBLE,
		 130, 2, 27, 14,
		 hwnd, (HMENU) ID_BUTTON, NULL, NULL);
  
  //up
  CreateWindowEx(0L, "BUTTON", "^",
		 WS_CHILD | WS_VISIBLE,
		 0, 40, 12, 12,
		 hwnd, (HMENU) ID_BUTTON_UP, NULL, NULL);
  
  //down
  CreateWindowEx(0L, "BUTTON", "V",
		 WS_CHILD | WS_VISIBLE,
		 0, 56+n_rows*13, 12, 12,
		 hwnd, (HMENU) ID_BUTTON_DN, NULL, NULL);
  
  //left
  CreateWindowEx(0L, "BUTTON", "<",
		 WS_CHILD | WS_VISIBLE,
		 13, 56+n_rows*13, 12, 12,
		 hwnd, (HMENU) ID_BUTTON_LF, NULL, NULL);
  
  //right
  CreateWindowEx(0L, "BUTTON", ">",
		 WS_CHILD | WS_VISIBLE,
		 135, 56+n_rows*13, 12, 12,
		 hwnd, (HMENU) ID_BUTTON_RT, NULL, NULL);


  //enter data
  CreateWindowEx(0L, "BUTTON", "O",
                 WS_CHILD | WS_VISIBLE,
                 35, 20, 10, 12,
                 hwnd, (HMENU) ID_BUTTON_EN, NULL, NULL);

  //del data
  CreateWindowEx(0L, "BUTTON", "X",
                 WS_CHILD | WS_VISIBLE,
                 45, 20, 10, 12,
                 hwnd, (HMENU) ID_BUTTON_DE, NULL, NULL);
      
  CreateWindowEx(0L, "BUTTON", "H",
                 WS_CHILD | WS_VISIBLE,
                 25, 20, 10, 12,
                 hwnd, (HMENU) ID_BUTTON_HM, NULL, NULL);

  //cursor id
  CreateWindowEx(0L, "STATIC", "?00:",
		 WS_CHILD | WS_VISIBLE,
		 0, 20, 25, 15,
		 hwnd, (HMENU) ID_STATIC_ID, NULL, NULL);
  
  InitXime(hwnd, XIME_NONE);	// init keyboard

  EnableXimeMinimize(TRUE);	// display minimize menu item
  
  SendMessage(hedit, EM_LIMITTEXT, 63, 0);
  SetFocus (hedit);	// default focus to edit control

  EnableMenuItem(hwnd->hMenu, IDM_SAVE, MF_DISABLED);

  ShowWindow (hwnd, iCmdShow);
  UpdateWindow (hwnd);

  while (GetMessage (&msg, NULL, 0, 0)){
    TranslateMessage (&msg);
    DispatchMessage (&msg);
  }      
  
  return msg.wParam;
}

LRESULT CALLBACK wproc (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam){
  HWND hctl;
  //  RECT rect;
  POINT pt;
  char m_buffer[2*256]="";
  int len=0;
  
  switch (iMsg){
  case WM_PAINT:
    SetWindowText(hwnd, caption);
    SendMessage(hwnd, WM_NCPAINT, 0, 0);
    MyPaint(hwnd);
    SetFocus (GetDlgItem (hwnd, ID_EDIT));
    break;
    
  case WM_COMMAND:	// got command from button
    if(last_cell)
      EnableMenuItem(hwnd->hMenu, IDM_SAVE, MF_ENABLED);      
    else
      EnableMenuItem(hwnd->hMenu, IDM_SAVE, MF_DISABLED);

    if ((LOWORD (wParam) == ID_BUTTON) &&
	(HIWORD (wParam) == BN_CLICKED))
      PostQuitMessage (0);

    if ((LOWORD (wParam) == ID_BUTTON_UP) &&
	(HIWORD (wParam) == BN_CLICKED) && m_row>0){
      enter_data(hwnd);--m_row;MyPaint(hwnd);}

    if ((LOWORD (wParam) == ID_BUTTON_DN) &&
	(HIWORD (wParam) == BN_CLICKED) && m_row<ROW_MAX-n_rows){
      enter_data(hwnd);++m_row;MyPaint(hwnd);}

    if ((LOWORD (wParam) == ID_BUTTON_LF) &&
	(HIWORD (wParam) == BN_CLICKED) && m_col>0){
      enter_data(hwnd);--m_col;MyPaint(hwnd);}

    if ((LOWORD (wParam) == ID_BUTTON_RT) &&
	(HIWORD (wParam) == BN_CLICKED) && m_col<COL_MAX-4){
      enter_data(hwnd);++m_col;MyPaint(hwnd);}

    if ((LOWORD (wParam) == ID_BUTTON_EN) &&
        (HIWORD (wParam) == BN_CLICKED) && m_col<COL_MAX-4){
      enter_data(hwnd);MyPaint(hwnd);}

    //if ((LOWORD (wParam) == ID_BUTTON_DE) &&
    //   (HIWORD (wParam) == BN_CLICKED) && m_col<COL_MAX-4){
    //  m_col=m_row=m_x=m_y=0;InvalidateRect (hwnd, NULL, TRUE);}

    if ((LOWORD (wParam) == ID_BUTTON_DE) &&
        (HIWORD (wParam) == BN_CLICKED) && m_col<COL_MAX-4){
      del_cell(hwnd);MyPaint(hwnd);}

    if ((LOWORD (wParam) == ID_BUTTON_HM) &&
        (HIWORD (wParam) == BN_CLICKED) && m_col<COL_MAX-4){
      SetDlgItemText(hwnd, ID_EDIT, "");
      m_col=m_row=m_x=m_y=0;MyPaint(hwnd);}

    switch(LOWORD (wParam)){
    case IDM_NEW:
      if( MessageBox(NULL,"Warning... \n \n"
		     "Content will be \ncleared, continue?", "New File",
		     MB_OKCANCEL | MB_ICONINFORMATION)==IDOK ){
	new_file();
	SetDlgItemText(hwnd, ID_EDIT,"");
	sprintf(caption, "XPRESO - New File");
      }
      break;

    case IDM_OPEN:
      if( MessageBox(NULL,"Warning... \n \n"
		     "Content will be \noverwritten, continue?", "Open File",
		     MB_OKCANCEL | MB_ICONINFORMATION)==IDOK ){
	new_file();
	SetDlgItemText(hwnd, ID_EDIT,"");

	DialogBoxEx(NULL, "DLG_OPEN_FILE", hwnd, LoadProc, dialog_info, 10);
      }
      
      break;

    case IDM_SAVE:
      DialogBoxEx(NULL, "DLG_SAVE_FILE", hwnd, SaveProc, dialog_info, 10);

      break;

    case IDM_CUT:
      len = GetWindowTextLength(GetDlgItem(hwnd, ID_EDIT));
      if(len>0){
	GetDlgItemText(hwnd, ID_EDIT, m_buffer, len + 1);
	WriteClipboard(m_buffer);
	del_cell(hwnd);
	InvalidateRect(hwnd, NULL, TRUE);
      }
      break;

    case IDM_COPY:
      len = GetWindowTextLength(GetDlgItem(hwnd, ID_EDIT));
      if(len>0){
	GetDlgItemText(hwnd, ID_EDIT, m_buffer, len + 1);
	WriteClipboard(m_buffer);
	InvalidateRect(hwnd, NULL, TRUE);
      }
      break;

    case IDM_PASTE:
      if( !IsClipboardEmpty()){
	ReadClipboard(m_buffer, 256*2);
	SetDlgItemText(hwnd, ID_EDIT, m_buffer);
	enter_data(hwnd);
	InvalidateRect(hwnd, NULL, TRUE);
      }
      break;

    case IDM_SORT:
      SetDlgItemText(hwnd, ID_EDIT, "");
      DialogBoxEx(NULL, "DLG_DO_SORT", hwnd, SortProc, dialog_info, 10);
      //SetDlgItemText(hwnd, ID_EDIT, "");
      break;

    case IDM_CHART:
      DialogBoxEx(NULL, "DLG_MK_CHART", hwnd, ChartProc, dialog_info, 10);
      //sprintf(m_range, "B01:B03");
      //sprintf(m_key,"A");
      //DialogBoxEx(NULL, "DLG_THE_CHART", hwnd, PieProc, dialog_info, 10);
      break;

    case IDM_JUMP:
      SetDlgItemText(hwnd, ID_EDIT, "");
      DialogBoxEx(NULL, "DLG_JUMP", hwnd, JumpProc, dialog_info, 10);

      //SetDlgItemText(hwnd, ID_EDIT, "");
      break;

    case IDM_HELP:
      //MessageBox(NULL,"Document is under \nconstruction","HELP",
      //	 MB_OK | MB_ICONINFORMATION);
      showhelp("xpreso","MAIN");
      break;

    case IDM_ABOUT:
      MessageBox(NULL,
		 "XPRESO\nVer. 0.9"
		 " \n \nSimple Spreadsheet\n \nAuthor: Y. Chao () "
		 "\nCopyright 2002.\nIssued under GPL.",
		 "About",
		 MB_OK | MB_ICONINFORMATION);
      
      showabout("xpreso");
      break;

    case IDM_CLOSE:
      if( MessageBox(NULL,"Warning... \n \n"
		     "Content will be gone,\ncontinue?", "Quit",
		     MB_OKCANCEL | MB_ICONINFORMATION)==IDOK )
	PostQuitMessage (0);

      break;

    }
    //InvalidateRect (hwnd, NULL, TRUE);
    break;
    
  case WM_SETFOCUS:	// re-direct focus to edit control
    SetFocus (GetDlgItem (hwnd, ID_EDIT));
    break;

  case WM_LBUTTONDOWN:
    pt.x=LOWORD(lParam);
    pt.y=HIWORD(lParam);

    m_old_x=m_old_y=0;
    if(pt.x>12 && pt.y>41 && pt.y<53){
      m_old_x=pt.x;
      m_old_y=pt.y;
    }
    break;

  case WM_LBUTTONUP:
    //    if (GetCapture() != hwnd)
    //  break;
    //ReleaseCapture();

    pt.x=LOWORD(lParam);
    pt.y=HIWORD(lParam);

    if(pt.x>12 && pt.y>41 && pt.y<53 && m_old_x!=0 && m_old_y!=0){
      //if(m_old_x!=0 && m_old_y!=0){
      n_chars+=(pt.x-m_old_x)/6;

      if(n_chars<6)
	n_chars=6;
      if(n_chars>24)
	n_chars=24;

      n_cols=148/(6*n_chars+1);

      m_old_x=m_old_y=0;

      fix_disp();
      MyPaint(hwnd);

      break;
    }

    if(pt.x>12 && pt.y>53 && pt.y<53+n_rows*13){
      enter_data(hwnd);

      m_x=(pt.x-12)/(6*n_chars+1);
      m_y=(pt.y-53)/13;
    
      //InvalidateRect (hwnd, NULL, TRUE);
      MyPaint(hwnd);
    }

    break;

  case WM_XIME_MINIMIZE:	// xime minimized
    n_rows=13;

    hctl = GetDlgItem (hwnd, ID_BUTTON_DN);
    //    GetWindowRect (hctl, &rect);
    SetWindowPos (hctl, 0,
		  0, 56+n_rows*13, 12, 12, 0);
    InvalidateRect (hctl, NULL, TRUE);
 
    hctl = GetDlgItem (hwnd, ID_BUTTON_LF);
    //    GetWindowRect (hctl, &rect);
    SetWindowPos (hctl, 0,
		  13, 56+n_rows*13, 12, 12, 0);
    InvalidateRect (hctl, NULL, TRUE);

    hctl = GetDlgItem (hwnd, ID_BUTTON_RT);
    //    GetWindowRect (hctl, &rect);
    SetWindowPos (hctl, 0,
		  135, 56+n_rows*13, 12, 12, 0);
    InvalidateRect (hctl, NULL, TRUE);

    if(m_row+n_rows>100){
      m_row-=(m_row+n_rows)-100;
    }

    //InvalidateRect (hwnd, NULL, TRUE);	// need repaint
    MyPaint(hwnd);

    break;

  case WM_XIME_RESTORE:	// xime restored
    n_rows=6;
    //    m_y=0;

    hctl = GetDlgItem (hwnd, ID_BUTTON_DN);
    //    GetWindowRect (hctl, &rect);
    SetWindowPos (hctl, 0,
		  0, 132, 12, 12, 0);

    hctl = GetDlgItem (hwnd, ID_BUTTON_LF);
    //    GetWindowRect (hctl, &rect);
    SetWindowPos (hctl, 0,
		  13, 132, 12, 12, 0);

    hctl = GetDlgItem (hwnd, ID_BUTTON_RT);
    //    GetWindowRect (hctl, &rect);
    SetWindowPos (hctl, 0,
		  148, 132, 12, 12, 0);

    if(m_y>5){
      m_row+=m_y-5;
      m_y=5;
    }

    //InvalidateRect (hwnd, NULL, TRUE);	// need repaint
    MyPaint(hwnd);

    break;

  case WM_KEYUP:
    if (wParam == VK_ESCAPE)	// ESCAPE pressed
      PostQuitMessage (0);
    break;
    
  default:
    return DefWindowProc (hwnd, iMsg, wParam, lParam);
  }      
  return 0;
}

void MyPaint (HWND hwnd){

  HDC hdc;
  PAINTSTRUCT ps;
//  HBRUSH hbrush, oldbrush;
  HPEN hpen, oldpen;
  COLORREF color_old;

  int i, j, len;
  char m_str_tmp[64];
  
  hdc = BeginPaint (hwnd, &ps);

  SelectObject (hdc, GetSystemFontObject ());

  //Write the cell ID under cursor
  sprintf(m_str_tmp, "%c%02d:", 'A'+m_col+m_x, m_row+m_y+1);
  SetDlgItemText(hwnd, ID_STATIC_ID, m_str_tmp);

  //printf("test\n");
  if(m_book[m_col+m_x][m_row+m_y]){
    if(m_book[m_col+m_x][m_row+m_y]->data[0]!='\''){
      sprintf(m_str_tmp, "Val=%-9.9g",
	      m_book[m_col+m_x][m_row+m_y]->value);
    }else{
      sprintf(m_str_tmp, "Val=%9.9s",
	      m_book[m_col+m_x][m_row+m_y]->data+1);
    }

    color_old=SetBkColor(hdc, GetSysColor(COLOR_3DLIGHT));
    TextOut(hdc, 30, 55+(n_rows)*13, m_str_tmp, -1);
    SetBkColor(hdc, color_old);
  }

  //Plot the axises
  for(i=0; i<=n_cols; i++){
    //for(i=0; i<=3; i++){
    sprintf(m_str_tmp,"   %c                    ",'A'+m_col+i);
    m_str_tmp[n_chars]='\0';
    TextOut (hdc, 13+i*(6*n_chars+1), 40, m_str_tmp, -1);
  }

  for(j=0; j<n_rows; j++){
    sprintf(m_str_tmp,"%02d", j+m_row+1);
    TextOut (hdc, 0, 53+j*13, m_str_tmp, -1);
  }

  //Write the entries
  len = GetWindowTextLength(GetDlgItem(hwnd, ID_EDIT));
  if( len ==0 ){  //avoid the bug if xime changes
    if(m_book[m_col+m_x][m_row+m_y])
      SetDlgItemText(hwnd, ID_EDIT, m_book[m_col+m_x][m_row+m_y]->data);
    else
      SetDlgItemText(hwnd, ID_EDIT, "");
  }

  //hpen = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_3DSHADOW));
  hpen = CreatePen(PS_SOLID, 1, RGB(255,255,255));
  oldpen = SelectObject(hdc, hpen);
  //hbrush = CreateSolidBrush(RGB(255,255,255));
  for(i=1; i<=n_cols; i++){
    MoveToEx(hdc, 12+i*(6*n_chars+1), 53, NULL);
    LineTo(hdc, 12+i*(6*n_chars+1), 52+n_rows*13);
  }
  for(j=1; j<n_rows+1; j++){
    MoveToEx(hdc, 13, 52+j*13, NULL);
    LineTo(hdc, 160, 52+j*13);
  }
  SelectObject(hdc, oldpen);
  DeleteObject(hpen);
  //DeleteObject(hbrush);

  //setbkmode(hdc, OPAQUE);
  color_old=SetBkColor(hdc, GetSysColor(COLOR_3DLIGHT));
  for(i=m_col; i<m_col+n_cols+1; i++){
    for(j=m_row; j<m_row+n_rows;j++){
      if(m_book[i][j] && m_book[i][j]->disp){
	//check right cell is empty
	if( m_book[i+1][j] ){
	  strncpy(m_str_tmp, m_book[i][j]->disp, n_chars);
	  m_str_tmp[n_chars]='\0';
	}else{
	  strcpy(m_str_tmp, m_book[i][j]->disp);
	  //m_str_tmp[n_chars*2]='\0';
	}
	TextOut (hdc, 13+(i-m_col)*(6*n_chars+1),
		 53+(j-m_row)*13, m_str_tmp, -1);
      }
    }
  }
	  
  SetBkColor(hdc, color_old);
  SetTextColor(hdc, GetSysColor(COLOR_CAPTIONTEXT));
  color_old=SetBkColor(hdc, GetSysColor(COLOR_3DSHADOW));
  if( !m_book[m_col+m_x][m_row+m_y] ){  //null
    strcpy(m_str_tmp, "                        ");
    m_str_tmp[n_chars]='\0';
    TextOut (hdc, 13+m_x*(6*n_chars+1), 53+m_y*13, m_str_tmp, -1);
  }else{
    sprintf(format_str, "%%-%d.%ds", n_chars, n_chars);
    sprintf(m_str_tmp, format_str,m_book[m_col+m_x][m_row+m_y]->disp);
    m_str_tmp[n_chars]='\0';
    TextOut (hdc, 13+m_x*(6*n_chars+1), 53+m_y*13, m_str_tmp, -1);
  }

  SetBkColor(hdc, color_old);

  EndPaint (hwnd, &ps);

  SetFocus (GetDlgItem (hwnd, ID_EDIT));
}

void enter_data(HWND hwnd){
  char buf[2*256];
  int len;

  len = GetWindowTextLength(GetDlgItem(hwnd, ID_EDIT));
  if(len>0){
    GetDlgItemText(hwnd, ID_EDIT, buf, len + 1);
    
    //check new entry
    if(!m_book[m_col+m_x][m_row+m_y]){
      m_book[m_col+m_x][m_row+m_y]=malloc(sizeof(cell));

      m_book[m_col+m_x][m_row+m_y]->prev=last_cell;
      if(first_cell){  //not the first one
	last_cell->next=m_book[m_col+m_x][m_row+m_y];
	last_cell=m_book[m_col+m_x][m_row+m_y];
      }else{
	first_cell=last_cell=m_book[m_col+m_x][m_row+m_y];
	m_book[m_col+m_x][m_row+m_y]->prev=NULL;
      }
      m_book[m_col+m_x][m_row+m_y]->next=NULL;
      m_book[m_col+m_x][m_row+m_y]->data[0]='\0';
      m_book[m_col+m_x][m_row+m_y]->disp[0]='\0';

      if(m_col_max<m_col+m_x)
	m_col_max=m_col+m_x;
      if(m_row_max<m_row+m_y)
	m_row_max=m_row+m_y;
    }else if(strcmp(m_book[m_col+m_x][m_row+m_y]->data, buf)==0){
      SetDlgItemText(hwnd, ID_EDIT, "");
      return;
    }
    
    strcpy(m_book[m_col+m_x][m_row+m_y]->data, buf);
    m_book[m_col+m_x][m_row+m_y]->id=(m_col+m_x)*256+(m_row+m_y);
    m_book[m_col+m_x][m_row+m_y]->value=0;

    //process the input here...
    if(buf[0]!='=' && is_text(buf, 0, strlen(buf))){      
      //force to be text
      if(buf[0]=='\''){
	if(strlen(buf)>1){
	  sprintf(format_str, "%%.%ds", n_chars*2);
	  sprintf(m_book[m_col+m_x][m_row+m_y]->disp, format_str, buf+1);
	}else
	  m_book[m_col+m_x][m_row+m_y]->disp[0]='\0';

	//m_book[m_col+m_x][m_row+m_y]->disp[n_chars*2]='\0';
      }else{
	sprintf(m_book[m_col+m_x][m_row+m_y]->data,"\'%s", buf);
	sprintf(format_str, "%%.%ds", n_chars*2);
	sprintf(m_book[m_col+m_x][m_row+m_y]->disp, format_str, buf);
	m_book[m_col+m_x][m_row+m_y]->value=0;
      }
    }else{
      //do evaluating here...
      if(buf[0]=='='){
	//replace refence to other cell as its value
	replace_cell(buf+1);

	m_book[m_col+m_x][m_row+m_y]->value=
	  eval_braket( buf+1, 0 , strlen(buf+1));
      }else{
	sscanf(buf, "%lf", &(m_book[m_col+m_x][m_row+m_y]->value));
      }

      sprintf(format_str, "%%%d.%dg", n_chars, n_chars-3);
      sprintf(m_book[m_col+m_x][m_row+m_y]->disp, format_str,
              m_book[m_col+m_x][m_row+m_y]->value);
      //walk-around for non-standard FPU
      if(log10(fabs(m_book[m_col+m_x][m_row+m_y]->value))>307)
        sprintf(m_book[m_col+m_x][m_row+m_y]->disp, "Err");

      fix_values();
    }
  }else if(m_book[m_col+m_x][m_row+m_y]){  //clear up
    m_book[m_col+m_x][m_row+m_y]->data[0]='\0';
    m_book[m_col+m_x][m_row+m_y]->disp[0]='\0';
    m_book[m_col+m_x][m_row+m_y]->value=0;
  }
  SetDlgItemText(hwnd, ID_EDIT, "");

}

void enter_data1(char *data){
  char buf[2*256];
  //int len;

  if(strlen(data)>0){
    strcpy(buf, data);
    //GetDlgItemText(hwnd, ID_EDIT, buf, len + 1);
    
    //check new entry
    if(!m_book[m_col+m_x][m_row+m_y]){
      m_book[m_col+m_x][m_row+m_y]=malloc(sizeof(cell));

      m_book[m_col+m_x][m_row+m_y]->prev=last_cell;
      if(first_cell){  //not the first one
	last_cell->next=m_book[m_col+m_x][m_row+m_y];
	last_cell=m_book[m_col+m_x][m_row+m_y];
      }else{
	first_cell=last_cell=m_book[m_col+m_x][m_row+m_y];
	m_book[m_col+m_x][m_row+m_y]->prev=NULL;
      }
      m_book[m_col+m_x][m_row+m_y]->next=NULL;
      m_book[m_col+m_x][m_row+m_y]->data[0]='\0';
      m_book[m_col+m_x][m_row+m_y]->disp[0]='\0';
      m_book[m_col+m_x][m_row+m_y]->value=0;

      if(m_col_max<m_col+m_x)
	m_col_max=m_col+m_x;
      if(m_row_max<m_row+m_y)
	m_row_max=m_row+m_y;
    }else if(strcmp(m_book[m_col+m_x][m_row+m_y]->data, buf)==0){
      //SetDlgItemText(hwnd, ID_EDIT, "");
      return;
    }
    
    strcpy(m_book[m_col+m_x][m_row+m_y]->data, buf);
    m_book[m_col+m_x][m_row+m_y]->id=(m_col+m_x)*256+(m_row+m_y);

    //process the input here...
    if(buf[0]!='=' && is_text(buf, 0, strlen(buf))){
      //content is text
      //check if force to be text
      if(buf[0]=='\''){
	if(strlen(buf)>1){
	  sprintf(format_str, "%%.%ds", n_chars*2);
	  sprintf(m_book[m_col+m_x][m_row+m_y]->disp, format_str, buf+1);
	}else
	  m_book[m_col+m_x][m_row+m_y]->disp[0]='\0';
	//m_book[m_col+m_x][m_row+m_y]->disp[n_chars*2]='\0';
      }else{
	sprintf(m_book[m_col+m_x][m_row+m_y]->data,"\'%s", buf);
	sprintf(format_str, "%%.%ds", n_chars*2);
	sprintf(m_book[m_col+m_x][m_row+m_y]->disp, format_str, buf);
      }
    }else{
      //numbers
      //do evaluating here...
      if(buf[0]=='='){
	//replace refence to other cell as its value
	replace_cell(buf+1);

	m_book[m_col+m_x][m_row+m_y]->value=
	  eval_braket( buf+1, 0 , strlen(buf+1));
      }else{
	sscanf(buf, "%lf", &(m_book[m_col+m_x][m_row+m_y]->value));
      }
      
      sprintf(format_str, "%%%d.%dg", n_chars, n_chars-3);
      sprintf(m_book[m_col+m_x][m_row+m_y]->disp, format_str,
              m_book[m_col+m_x][m_row+m_y]->value);
      if(log10(fabs(m_book[m_col+m_x][m_row+m_y]->value))>307)
        sprintf(m_book[m_col+m_x][m_row+m_y]->disp, "Err");
    }
  }else if(m_book[m_col+m_x][m_row+m_y]){  //clear up
    m_book[m_col+m_x][m_row+m_y]->data[0]='\0';
    m_book[m_col+m_x][m_row+m_y]->disp[0]='\0';
    m_book[m_col+m_x][m_row+m_y]->value=0;
  }
}

void del_cell(HWND hwnd){
  cell* ptr;

  ptr=m_book[m_col+m_x][m_row+m_y];

  if(ptr){  //cell exists
    if(ptr->prev)
      ptr->prev->next=ptr->next;
    else
      first_cell=ptr->next;

    if(ptr->next)
      ptr->next->prev=ptr->prev;
    else
      last_cell=ptr->prev;

    free(ptr);

    m_book[m_col+m_x][m_row+m_y]=NULL;

  }
  SetDlgItemText(hwnd, ID_EDIT, "");

}

void fix_values(){
  int i;
  char buf[256];
  cell* ptr;

  //reference level max: 3
  for(i=0; i<3; i++){
    for(ptr=first_cell; ptr!=NULL; ptr=ptr->next){
      //check if cell has reference
      if(ptr->data && ptr->data[0]=='='){
	strcpy(buf, ptr->data);
	replace_cell(buf+1);

	ptr->value=eval_braket(buf+1, 0, strlen(buf+1));
	
	sprintf(format_str, "%%%d.%dg", n_chars, n_chars-3);
	sprintf(ptr->disp, format_str,ptr->value);
	if(log10(fabs(ptr->value))>307)
	  sprintf(ptr->disp, "Err");
      }
    }
  }
}

void fix_disp(){
  cell* ptr;

  for(ptr=first_cell; ptr!=NULL; ptr=ptr->next){
    //check the data here...
    if(ptr->data[0]=='\''){   //text
      if(strlen(ptr->data)>1){
	sprintf(format_str, "%%.%ds", n_chars*2);
	sprintf(ptr->disp, format_str, ptr->data+1);
      }else
	ptr->disp[0]='\0';

    }else{   //numbers
      sprintf(format_str, "%%%d.%dg", n_chars, n_chars-3);
      sprintf(ptr->disp, format_str, ptr->value);

      //walk-around for non-standard FPU
      if(log10(fabs(ptr->value))>307)
        sprintf(ptr->disp, "Err");

    }
  }
}

void new_file(){
  int m_id_x=0, m_id_y=0;
  cell* ptr;

  for(ptr=first_cell; ptr!=NULL; ptr=ptr->next){
    m_id_x=ptr->id/256;
    m_id_y=ptr->id%256;
    if(m_book[m_id_x][m_id_y])
      m_book[m_id_x][m_id_y]=NULL;

    free(ptr);
  }

  m_col_max=m_row_max=0;
  m_x=m_y=0;
  first_cell=last_cell=NULL;
}

void load_file(FILE* file){
  char m_buff[256*5]="";
  int m_start, m_end, len;
  int m_quote=0;

  m_col=m_row=m_x=m_y=0;
  fgets(m_buff, 256*5-1, file);
  len=strlen(m_buff);
  m_buff[len-1]='\0';  //remove the char NL
  while(!feof(file)&&m_row<100){
    m_start=0;
    m_quote=0;
    for(m_end=0; m_end<=len; m_end++){
      if(m_buff[m_start]=='"'){
        //m_start++;
        m_quote++;
      }

      if((m_buff[m_end]==','&& m_buff[m_end-1]!='"') ||
	 (m_buff[m_end]==','&& m_buff[m_end-1]=='"' && m_buff[m_start]=='"')
	 || m_buff[m_end]=='\0'|| m_buff[m_end]=='\n'){
	if((m_end-m_start)>0){
	  m_buff[m_end]='\0';
	  if(m_buff[m_start]=='"' && m_buff[m_end-1]=='"'){
	    m_start++;
	    m_buff[m_end-1]='\0';
	  }
	  enter_data1(m_buff+m_start);
	    m_quote=0;
	}//else
	m_start=m_end+1;
	m_col++;
      }
    }
    m_col=0;
    m_row++;
    fgets(m_buff, 256*5-1, file);
    len=strlen(m_buff);
    m_buff[len-1]='\0';  //remove the char NL
  }
  m_col=m_row=m_x=m_y=0;

  //fix ref. cells here!
  fix_values();
}

void save_file(FILE* file){
  int m_id_x=0,m_id_y=0;
  cell* ptr;
  int i, j;

  m_col_max=m_row_max=0;
  for(ptr=first_cell; ptr!=NULL; ptr=ptr->next){
    m_id_x=ptr->id/256;
    m_id_y=ptr->id%256;
    if(m_book[m_id_x][m_id_y]){
      if(m_col_max<m_id_x) m_col_max=m_id_x;
      if(m_row_max<m_id_y) m_row_max=m_id_y;
    }
  }

  for(j=0; j<=m_row_max; j++){
    for(i=0; i<m_col_max; i++){
      if(m_book[i][j])
	fprintf(file, "%s,", m_book[i][j]->data);
      else
	fprintf(file, ",");
    }
    if(m_book[i][j])
      fprintf(file, "%s\n", m_book[i][j]->data);
    else
      fprintf(file, "\n");
  }

}


void replace_cell(char* buf){
  char m_buf[256*2]="";
  char m_buf_tmp[256*2]="";
  char *ptr;
  int m_id_x, m_id_y;
  int m_x1, m_x2, m_y1, m_y2;
  double m_sum;
  int i, j;

  while((ptr=strstr(buf,"sum("))!=NULL || (ptr=strstr(buf,"SUM("))!=NULL){
    if((ptr[4]>='A' && ptr[4]<='Z')||(ptr[4]>='a' && ptr[4]<='z')){
      if(ptr[5]>='0' && ptr[5]<='9' && ptr[5]>='0' && ptr[5]<='9'){
	if(ptr[4]>='A' && ptr[4]<='Z')
	  m_x1=ptr[4]-'A';
	else
	  m_x1=ptr[4]-'a';
	if(ptr[8]>='A' && ptr[8]<='Z')
	  m_x2=ptr[8]-'A';
	else
	  m_x2=ptr[8]-'a';
	m_y1=(ptr[5]-'0')*10+ptr[6]-'0'-1;
	m_y2=(ptr[9]-'0')*10+ptr[10]-'0'-1;

	m_sum=0;
	for(i=m_x1; i<=m_x2; i++){
	  for(j=m_y1; j<=m_y2; j++){
	    if(m_book[i][j])
	      m_sum+=m_book[i][j]->value;
	  }
	}
	sprintf(m_buf, "%f%s", m_sum, ptr+12);
	strcpy(ptr, m_buf);
      }
    }
  }

  m_buf[0]=m_buf_tmp[0]='\0';
  for(i=0; i< strlen(buf); i++){
    if((buf[i]>='A' && buf[i]<='Z')||(buf[i]>='a' && buf[i]<='z')){
      if(buf[i+1]>='0' && buf[i+1]<='9' && buf[i+2]>='0' && buf[i+2]<='9'){
	if(buf[i]>='A' && buf[i]<='Z')
	  m_id_x=buf[i]-'A';
	else
	  m_id_x=buf[i]-'a';
	m_id_y=(buf[i+1]-'0')*10+buf[i+2]-'0'-1;
	if(m_book[m_id_x][m_id_y]){
	  sprintf(format_str, "%%%d.%dg", n_chars, n_chars-3);
	  sprintf(m_buf_tmp, format_str, m_book[m_id_x][m_id_y]->value);
	}
	strcat(m_buf, m_buf_tmp);
	i+=2;
      }else
	strncat(m_buf, buf+i, 1);
    }else{
      strncat(m_buf, buf+i, 1);
    }
  }
  strcpy(buf, m_buf);
}

int is_text(char str[], int m_begin, int m_end){
  int i=0;

  if((str[0]=='-' || str[0]=='+') && m_end-m_begin>1)
    m_begin++;

  for(i=m_begin; i< m_end; i++){
    if((str[i]<'0' || str[i]>'9')&&str[i]!='.')
      return(1);
  }
  return (0);
}

static LRESULT CALLBACK LoadProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){

  int iIndex;
  char buf[128], *ptr;
  DIR *dir;
  struct dirent *entry;

  switch ( message ){
  case WM_INITDIALOG:
    //sprintf(buf, "%s/data/ebk", getenv("PDAROOT"));
    sprintf(buf, "/pda/gv/data/ebk");

    if ((dir = opendir(buf))==NULL){
      MessageBox (NULL, buf, "Opendir Failed!!",MB_OK | MB_ICONERROR);
      EndDialog (hDlg, IDCANCEL);
      break;
    }

    while ((entry = readdir(dir)) != NULL){
      // filter
      ptr = strstr(entry->d_name, ".txt");
      if (! ptr) ptr = strstr(entry->d_name, ".TXT");
      if (! ptr) ptr = strstr(entry->d_name, ".Txt");
      if (! ptr) ptr = strstr(entry->d_name, ".csv");
      if (! ptr) ptr = strstr(entry->d_name, ".CSV");
      if (! ptr) ptr = strstr(entry->d_name, ".Csv");
      if ( ptr && (ptr[4]=='\0')){
	// add to list
	SendMessage (GetDlgItem (hDlg, IDC_LIST), LB_ADDSTRING, 0,
		     (LPARAM) (LPSTR) entry->d_name);
      }
    }

    // close dir
    closedir (dir);

    SetFocus(GetDlgItem(hDlg, IDC_LIST));
    //SendMessage(hDlg, WM_COMMAND, MAKELONG(IDC_TYPE,CBN_SELCHANGE), 0);

    break;

  case WM_KEYDOWN:
    switch( LOWORD( wParam ) ){
    case VK_ESCAPE:
      SendMessage(hDlg, WM_COMMAND, MAKELONG(IDCANCEL, BN_CLICKED), 0);
      break;
    default:
      return DefWindowProc(hDlg, message, wParam, lParam);
    }
    return 0;
    
    
  case WM_COMMAND:
    switch ( LOWORD( wParam ) ){      
    case IDC_DEL:
      if ((iIndex = SendMessage(GetDlgItem(hDlg, IDC_LIST), 
				LB_GETCURSEL, 0, 0)) == LB_ERR)
	return TRUE;
      
      SendMessage (GetDlgItem(hDlg, IDC_LIST), LB_GETTEXT,
		   iIndex, (LPARAM) filename);

      if (MessageBox (hDlg, "Really want to delete?", NULL,
		      MB_YESNO | MB_ICONQUESTION) == IDYES){
	  sprintf(fullname, "%s/data/ebk/%s", getenv("PDAROOT"), filename);
	  unlink (fullname);
	  SendMessage (GetDlgItem(hDlg, IDC_LIST), LB_DELETESTRING, iIndex, 0);
	}
      break;
      
    case IDC_OK:
      /* Get file name */
      if ((iIndex = SendMessage(GetDlgItem(hDlg, IDC_LIST),
				LB_GETCURSEL, 0, 0)) == LB_ERR)
	return TRUE;
      SendMessage(GetDlgItem(hDlg, IDC_LIST), LB_GETTEXT,
		  iIndex, (LPARAM) filename);

      sprintf(fullname, "%s/data/ebk/%s", getenv("PDAROOT"), filename);
      
      data_file=fopen(fullname,"r");
      if(data_file){
	load_file(data_file);
	fclose(data_file);
	ptr=strstr(filename, ".csv");
	if(ptr)
	  ptr[0]='\0';
	else{
	  ptr=strstr(filename, ".txt");
	  ptr[0]='\0';
	}

	sprintf(caption, "XPRESO - %.8s", filename);

	n_rows=13;
	selectXime(XIME_NONE);
      }else{
	MessageBox(NULL,"Error opening\n data file!!!","Error",
		   MB_OK | MB_ICONERROR);
      }

      EndDialog (hDlg, IDCANCEL);
      return TRUE;

      break;

    case IDC_CANCEL:
      EndDialog (hDlg, IDCANCEL);
      return FALSE;

    default:
      return FALSE;
    }
  }
  
  return DefWindowProc( hDlg, message, wParam, lParam );
}

static LRESULT CALLBACK SaveProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){
  static HWND hwndOld;

  HWND	hedit;
  int	iIndex;
  int	nlen=0;
  char buf[128], *ptr;
  DIR *dir;
  struct dirent *entry;

  switch ( message ){
  case WM_INITDIALOG:
    selectXime (XIME_ASCII);
    hwndOld = (HWND) setXime (hDlg);
    sprintf(buf, "%s/data/ebk", getenv("PDAROOT"));

    if ((dir = opendir(buf))==NULL){
      MessageBox (NULL, buf, "Opendir Failed!!",MB_OK | MB_ICONERROR);
      EndDialog (hDlg, IDCANCEL);
      break;
    }

    while ((entry = readdir(dir)) != NULL){
      // filter
      ptr = strstr (entry->d_name, ".txt");
      //if (! ptr) ptr = strstr (entry->d_name, ".TXT");
      //if (! ptr) ptr = strstr (entry->d_name, ".Txt");
      if ( ptr && (ptr[4]=='\0')){
	// add to list
	SendMessage(GetDlgItem (hDlg, IDC_LIST), LB_ADDSTRING, 0,
		     (LPARAM) (LPSTR) entry->d_name);
      }
    }

    // close dir
    closedir (dir);

    SetFocus(GetDlgItem(hDlg, IDC_FILE));

    SendMessage(GetDlgItem(hDlg, IDC_FILE), EM_SETLIMITTEXT, 127, 0);
    if(strstr(filename, ".csv")==NULL)
      strcat(filename, ".csv");
    SetWindowText(GetDlgItem(hDlg, IDC_FILE), filename);
    break;

  case WM_PAINT:
    InvalidateRect(GetDlgItem (hDlg, IDC_FILE), NULL, TRUE);
    InvalidateRect(GetDlgItem (hDlg, IDC_LIST), NULL, TRUE);
    SetFocus(GetDlgItem (hDlg, IDC_FILE));
    break;

  case WM_SETFOCUS:	// re-direct focus to edit control
    SetFocus (GetDlgItem (hDlg, IDC_FILE));
    break;
    
  case WM_COMMAND:
    switch ( LOWORD( wParam ) ){      
    case IDC_LIST:
      if (HIWORD(wParam) == LBN_SELCHANGE){
	iIndex = SendMessage(GetDlgItem(hDlg, IDC_LIST), LB_GETCURSEL, 0, 0);
	SendMessage(GetDlgItem(hDlg, IDC_LIST), LB_GETTEXT,
		    iIndex, (LPARAM)buf);
	SetWindowText (GetDlgItem(hDlg, IDC_FILE), buf);
	SetFocus (GetDlgItem (hDlg, IDC_FILE));
      }
      break;
      
    case IDC_OK:
      // Get file name
      hedit = GetDlgItem(hDlg, IDC_FILE);
      nlen = GetWindowText(hedit, filename, 127);
      if (nlen <= 0) {
	SetFocus (hedit);
	return TRUE;
      }
      sprintf(fullname, "%s/data/ebk/%s", getenv("PDAROOT"), filename);

      if(fopen(fullname,"r") &&
	 MessageBox(hDlg, "File exists,\n continue?", "Warning",
		    MB_YESNO | MB_ICONQUESTION) == IDNO) {
	SetFocus (GetDlgItem (hDlg, IDC_FILE));
	return 0;
      }
      if((data_file=fopen(fullname,"w"))!=NULL){
	save_file(data_file);
        fclose(data_file);
	ptr=strstr(filename, ".csv");
	if(ptr)
	  ptr[0]='\0';
	else{
	  ptr=strstr(filename, ".txt");
	  ptr[0]='\0';
	}
	sprintf(caption, "XPRESO - %s", filename);
      }else{
         MessageBox(NULL,"Error opening\n data file!!!","Error",
                    MB_OK | MB_ICONERROR);
      }

      setXime (hwndOld);
      EndDialog (hDlg, IDOK);
      return TRUE;
      
    case IDC_CANCEL:
      setXime (hwndOld);
      //ShowWindow (hXime, SW_HIDE);
      SetDlgItemText(hwndOld, ID_EDIT, "");
      EndDialog (hDlg, IDCANCEL);
      return IDCANCEL;
    }
    return 0;
  }
  
  return DefWindowProc( hDlg, message, wParam, lParam );
}

static LRESULT CALLBACK SortProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){
  static HWND hwndOld;

  HWND	hedit;
  int nlen=0, m_order=0;
  int m_min_x, m_min_y;
  int m_max_x, m_max_y;
  int m_comp;
  int i, j, k;
  struct cell *tmp;

  switch ( message ){
  case WM_INITDIALOG:
    selectXime (XIME_ASCII);
    hwndOld = (HWND) setXime (hDlg);
    SetDlgItemText(hDlg, IDC_RANGE, m_range);
    SetDlgItemText(hDlg, IDC_KEY, m_key);
    SendMessage(GetDlgItem (hDlg, IDC_RANGE), EM_LIMITTEXT, 10, 0);
    SendMessage (GetDlgItem (hDlg, IDC_RADIO_ASC), BM_SETCHECK, 1, 0);
    SetFocus(GetDlgItem (hDlg, IDC_RANGE));

    break;

  case WM_PAINT:
    //InvalidateRect(GetDlgItem (hDlg, IDC_FILE), NULL, TRUE);
    //InvalidateRect(GetDlgItem (hDlg, IDC_LIST), NULL, TRUE);
    //SetFocus(GetDlgItem (hDlg, IDC_RANGE));
    break;

  case WM_COMMAND:
    switch ( LOWORD( wParam ) ){      
    case IDC_OK:
      // Get file name
      hedit = GetDlgItem(hDlg, IDC_RANGE);
      nlen = GetWindowText(hedit, m_range, 8);
      if (nlen <= 0) {
	SetFocus (hedit);
	return TRUE;
      }
      hedit = GetDlgItem(hDlg, IDC_KEY);
      nlen = GetWindowText(hedit, m_key, 4);
      if (nlen <= 0) {
	SetFocus (hedit);
	return TRUE;
      }

      if(SendMessage(GetDlgItem(hDlg, IDC_RADIO_ASC),
		     BM_GETCHECK, 0, 0)==BST_CHECKED)
	m_order=1;
      else
	m_order=-1;

      if(m_range[0]>='A' && m_range[0]<='Z')
	m_min_x=m_range[0]-'A';
      else
	m_min_x=m_range[0]-'a';

      if(m_key[0]>='A' && m_key[0]<='Z')
	m_comp=m_key[0]-'A';
      else
	m_comp=m_key[0]-'a';

      if(m_range[4]>='A' && m_range[4]<='Z')
	m_max_x=m_range[4]-'A';
      else
	m_max_x=m_range[4]-'a';
   
      m_min_y=(m_range[1]-'0')*10+(m_range[2]-'0')-1;
      m_max_y=(m_range[5]-'0')*10+(m_range[6]-'0')-1;

      if(m_min_x>=0&&m_min_x<27&&m_max_x>=0&&m_max_x<27&&m_min_x<=m_max_x&&
	 m_min_y>=0&&m_min_y<100&&m_max_y>=0&&m_max_y<100&&m_min_y<=m_max_y &&
	 m_comp>=m_min_x&&m_comp<=m_max_x){

	//shell sort
	for(i=m_min_y; i<m_max_y; i++){
	  for(j=i+1; j<=m_max_y; j++){
	    if(m_book[m_comp][i] && m_book[m_comp][j]){
	      if((m_book[m_comp][i]->value-
		  m_book[m_comp][j]->value)*m_order>0){
		//compared based on value, swap two rows
		for(k=m_min_x; k<=m_max_x; k++){
		  tmp=m_book[k][i];
		  m_book[k][i]=m_book[k][j];
		  m_book[k][j]=tmp;
		  if(m_book[k][i]) m_book[k][i]->id=k*256+i;
		  if(m_book[k][j]) m_book[k][j]->id=k*256+j;
		  if(m_col+m_x==k&&m_row+m_y==i){ m_y=j-m_row;
		  }else if(m_col+m_x==k&&m_row+m_y==j) m_y=i-m_row;
		}
	      }else if(strcmp(m_book[m_comp][i]->data,
			      m_book[m_comp][j]->data)*m_order>0){
		//compared based on text, swap two rows
		for(k=m_min_x; k<=m_max_x; k++){
		  tmp=m_book[k][i];
		  m_book[k][i]=m_book[k][j];
		  m_book[k][j]=tmp;
		  if(m_book[k][i]) m_book[k][i]->id=k*256+i;
		  if(m_book[k][j]) m_book[k][j]->id=k*256+j;
		  if(m_col+m_x==k&&m_row+m_y==i){ m_y=j-m_row;
		  }else if(m_col+m_x==k&&m_row+m_y==j) m_y=i-m_row;
		}
	      }
	    }else{
	      //emtpy cell, treat as min, swap anyway
	      printf("empty cell...\n");
	      if(m_book[m_comp][j]&&m_book[m_comp][j]->data[0]!='\0'){
		for(k=m_min_x; k<=m_max_x; k++){
		  tmp=m_book[k][i];
		  m_book[k][i]=m_book[k][j];
		  m_book[k][j]=tmp;
		  if(m_book[k][i]) m_book[k][i]->id=k*256+i;
		  if(m_book[k][j]) m_book[k][j]->id=k*256+j;
		  if(m_col+m_x==k&&m_row+m_y==i){ m_y=j-m_row;
		  }else if(m_col+m_x==k&&m_row+m_y==j) m_y=i-m_row;
		}
	      }
	    }
	  }
	}

	setXime (hwndOld);
	EndDialog (hDlg, IDOK);
	return TRUE;
      }else{
	SetFocus (GetDlgItem(hDlg, IDC_RANGE));
	return TRUE;
      }
      return TRUE;
      
    case IDC_CANCEL:
      setXime (hwndOld);
      //ShowWindow (hXime, SW_HIDE);
      SetDlgItemText(hwndOld, ID_EDIT, "");
      EndDialog (hDlg, IDCANCEL);
      return IDCANCEL;
    }
    return 0;
  }
  
  return DefWindowProc( hDlg, message, wParam, lParam );
}

static LRESULT CALLBACK ChartProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){
  static HWND hwndOld;

  HWND	hedit;
  int	nlen=0;

  switch ( message ){
  case WM_INITDIALOG:
    selectXime (XIME_ASCII);
    hwndOld = (HWND) setXime (hDlg);
    SetDlgItemText(hDlg, IDC_RANGE, m_range);
    SetDlgItemText(hDlg, IDC_KEY, m_key);
    SendMessage(GetDlgItem(hDlg, IDC_RANGE), EM_LIMITTEXT, 10, 0);
    SendMessage(GetDlgItem(hDlg, IDC_RADIO_PIE), BM_SETCHECK, 1, 0);
    //EnableWindow(GetDlgItem(hDlg, IDC_RADIO_LINE),0);
    //EnableWindow(GetDlgItem(hDlg, IDC_RADIO_BAR),0);
    SetFocus (GetDlgItem (hDlg, IDC_RANGE));

    break;

  case WM_PAINT:
    //SetFocus(GetDlgItem (hDlg, IDC_RANGE));
    break;

  case WM_COMMAND:
    switch ( LOWORD( wParam ) ){      
    case IDC_OK:
      // Get file name
      hedit = GetDlgItem(hDlg, IDC_RANGE);
      nlen = GetWindowText(hedit, m_range, 8);
      if (nlen <= 0) {
	SetFocus (hedit);
	return TRUE;
      }
      hedit = GetDlgItem(hDlg, IDC_KEY);
      nlen = GetWindowText(hedit, m_key, 4);
      if (nlen <= 0) {
	SetFocus (hedit);
	return TRUE;
      }

      if(SendMessage(GetDlgItem(hDlg, IDC_RADIO_PIE),
		     BM_GETCHECK, 0, 0)==BST_CHECKED){
	DialogBoxEx(NULL, "DLG_PIE_CHART", hDlg, PieProc, dialog_info, 10);
      }else if(SendMessage(GetDlgItem(hDlg, IDC_RADIO_LINE),
		     BM_GETCHECK, 0, 0)==BST_CHECKED){
	DialogBoxEx(NULL, "DLG_LINE_CHART", hDlg, LineProc, dialog_info, 10);
      }else if(SendMessage(GetDlgItem(hDlg, IDC_RADIO_BAR),
		     BM_GETCHECK, 0, 0)==BST_CHECKED){
	DialogBoxEx(NULL, "DLG_BAR_CHART", hDlg, BarProc, dialog_info, 10);
      }
      //setXime (hwndOld);
      //EndDialog (hDlg, IDOK);
      return TRUE;
      
    case IDC_CANCEL:
      setXime (hwndOld);
      //ShowWindow (hXime, SW_HIDE);
      SetDlgItemText(hwndOld, ID_EDIT, "");
      EndDialog (hDlg, IDCANCEL);
      return IDCANCEL;
    }
    return 0;
  }
  
  return DefWindowProc( hDlg, message, wParam, lParam );
}

static LRESULT CALLBACK PieProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){

  HDC hdc;
  PAINTSTRUCT ps;
  HBRUSH hbrush, oldbrush;
  HPEN hpen, oldpen;
  COLORREF color_old;

  int m_min_x, m_min_y;
  int m_max_x, m_max_y;
  int m_symbol;

  int i, j;
  double m_total;
  double m_ang_sum, m_ang_det;

  char buf[128];

  switch ( message ){
  case WM_INITDIALOG:
    SetWindowText(hDlg, "PIE");
#if GVPDA
    SetScreenMode(4);      // 4bpp
    ResetScreen();
#else
    for(i=0; i<16; i++){
      pals[i].peRed=16*i;
      pals[i].peGreen=16*i;
      pals[i].peBlue=16*i;
    }
#endif
    break;

  case WM_PAINT:
    hdc = BeginPaint (hDlg, &ps);
    
    if(m_range[0]>='A' && m_range[0]<='Z')
      m_min_x=m_range[0]-'A';
    else
      m_min_x=m_range[0]-'a';

    if(m_key[0]>='A' && m_key[0]<='Z')
      m_symbol=m_key[0]-'A';
    else
      m_symbol=m_key[0]-'a';

    //Only single column allowed
    if(m_range[4]>='A' && m_range[4]<='Z')
      m_max_x=m_range[4]-'A';
    else
      m_max_x=m_range[4]-'a';
   
    m_min_y=(m_range[1]-'0')*10+(m_range[2]-'0')-1;
    m_max_y=(m_range[5]-'0')*10+(m_range[6]-'0')-1;

    if( m_min_x>=0&&m_min_x<27 &&
	m_min_y>=0&&m_min_y<100&&m_max_y>=0&&m_max_y<100 && m_min_y<=m_max_y){

      m_total=0;
      for(i=m_min_y; i<=m_max_y; i++){
	if(m_book[m_min_x][i]){
	  m_total+=fabs(m_book[m_min_x][i]->value);
	}
      }

      GetSystemPaletteEntries(hdc, 0, 16, (LPPALETTEENTRY)&pals);
      hpen = CreatePen(PS_SOLID, 1, RGB(0,0,0));
      oldpen = SelectObject(hdc, hpen);
      hbrush = CreateSolidBrush(RGB(255,255,255));
      oldbrush = SelectObject(hdc, hbrush);
      //SelectObject (hdc, (HPEN)GetStockObject(BLACK_PEN));
      color_old=SetBkColor(hdc, GetSysColor(COLOR_3DLIGHT));
 
      m_ang_sum=0;
      for(i=m_min_y; i<=m_max_y; i++){
	if(m_book[m_min_x][i]&&fabs(m_book[m_min_x][i]->value)>0){
	  m_ang_det=fabs(m_book[m_min_x][i]->value)/m_total*3.141592653*2;
	  if(i%2==0){
	    hbrush=CreateSolidBrush(RGB(pals[i%16].peRed,
					pals[i%16].peGreen,pals[i%16].peBlue));
	  }else{
	    hbrush=CreateSolidBrush(RGB(pals[(i+8)%16].peRed,
					pals[(i+8)%16].peGreen,
					pals[(i+8)%16].peBlue));
	  }
	  SelectObject(hdc, hbrush);
	  if(m_ang_det>0.1){
	    Pie(hdc,16,16, 144, 144,cos(m_ang_sum)*64+.5,-sin(m_ang_sum)*64+.5,
	      cos(m_ang_sum+m_ang_det)*64+.5,-sin(m_ang_sum+m_ang_det)*64+.5);
	  }// too small area will cause problem...
	  m_ang_sum+=m_ang_det;
	}
      }

      //SelectObject(hdc, (HPEN)GetStockObject(WHITE_PEN));
      m_ang_sum=0;
      for(i=m_min_y; i<=m_max_y; i++){
	if(m_book[m_min_x][i]&&fabs(m_book[m_min_x][i]->value)>0){
	  m_ang_det=fabs(m_book[m_min_x][i]->value)/m_total*3.141592653*2;
	  m_ang_sum+=m_ang_det;
	  //MoveToEx(hdc,80,79,NULL);
	  //LineTo(hdc,80+cos(m_ang_sum)*63+.5,79-sin(m_ang_sum)*63+.5);
	  SelectObject (hdc, GetSystemFontObject());
	  sprintf(buf,"%-3.3g",m_book[m_min_x][i]->value);
	  TextOut (hdc, 80+cos(m_ang_sum-m_ang_det/2)*64-2*6,
	  	   80-sin(m_ang_sum-m_ang_det/2)*64-12, buf, -1);
	  sprintf(buf,"(%2.1f%%)",fabs(m_book[m_min_x][i]->value)/m_total*100);
	  TextOut (hdc, 80+cos(m_ang_sum-m_ang_det/2)*64-3*6,
	  	   80-sin(m_ang_sum-m_ang_det/2)*64, buf, -1);
	}
      }

      //SelectObject (hdc, (HPEN)GetStockObject(BLACK_PEN));
      SetBkMode(hdc, TRANSPARENT);
      TextOut(hdc, 2, 148, "Symbol:", -1);
      MoveToEx(hdc,0,160,NULL);
      LineTo(hdc,160,160);
      MoveToEx(hdc,79,160,NULL);
      LineTo(hdc,79,240);

      //show the symbol keys...
      j=0;
      for(i=m_min_y; i<=m_max_y&& j<m_min_y+10; i++){
	if(m_book[m_min_x][i]){
	  if(i%2==0){
	    hbrush=CreateSolidBrush(RGB(pals[i%16].peRed,
					pals[i%16].peGreen,pals[i%16].peBlue));
	  }else{
	    hbrush=CreateSolidBrush(RGB(pals[(i+8)%16].peRed,
					pals[(i+8)%16].peGreen,
					pals[(i+8)%16].peBlue));
	  }
	  SelectObject(hdc, hbrush);
	  Rectangle(hdc,2+(int)(j/5)*80,167+j%5*15,
		    ((int)j/5)*80+9,173+j%5*15);
	  sprintf(buf, "%4.1f%%",fabs(m_book[m_min_x][i]->value)/m_total*100);
	  if(m_book[m_symbol][i]){
	    sprintf(buf, "%4.1f%%%-6.6s",
		    fabs(m_book[m_min_x][i]->value)/m_total*100,
		    m_book[m_symbol][i]->disp);
	  }
	  SelectObject (hdc, GetSystemFontObject());
	  TextOut(hdc, 10+(int)(j/5)*80, 165+j%5*15, buf, -1);
	  j++;
	}
      }

      SelectObject(hdc, oldpen);
      SelectObject(hdc, oldbrush);
      DeleteObject(hpen);
      DeleteObject(hbrush);

      SetBkColor(hdc, color_old);
    }else{
      //MessageBox(NULL,"Input range error...", "Error", MB_OK | MB_ICONERROR);
      EndPaint (hDlg, &ps);
      EndDialog (hDlg, IDOK);
      return TRUE;
    }
    
    //ReleaseDC(hDlg,hdc);
    EndPaint (hDlg, &ps);

    //InvalidateRect(GetDlgItem (hDlg, IDC_FILE), NULL, TRUE);
    //InvalidateRect(GetDlgItem (hDlg, IDC_LIST), NULL, TRUE);
    //SetFocus(GetDlgItem (hDlg, IDC_RANGE));
    break;

  case WM_COMMAND:
    switch ( LOWORD( wParam ) ){      
    case IDC_OK:
      EndDialog (hDlg, IDOK);
      return TRUE;
      break;    
    }
  default:
    return DefWindowProc( hDlg, message, wParam, lParam );
  }
  
  return 0;
}

static LRESULT CALLBACK LineProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){

  HDC hdc;
  PAINTSTRUCT ps;
  HBRUSH hbrush, oldbrush;
  HPEN hpen, oldpen;
  COLORREF color_old;

  int m_min_x, m_min_y;
  int m_max_x, m_max_y;
  int m_symbol;

  int i, j;
  //double m_axis_max, m_axis_min; // max & min value of Y-axis
  double m_val_max, m_val_min, m_scale;

  char buf[128];

  switch ( message ){
  case WM_INITDIALOG:
    SetWindowText(hDlg, "Line");
#if GVPDA
    SetScreenMode(4);      // 4bpp
    ResetScreen();
#else
    for(i=0; i<16; i++){
      pals[i].peRed=16*i;
      pals[i].peGreen=16*i;
      pals[i].peBlue=16*i;
    }
#endif
    break;

  case WM_PAINT:
    hdc = BeginPaint (hDlg, &ps);
    
    if(m_range[0]>='A' && m_range[0]<='Z')
      m_min_x=m_range[0]-'A';
    else
      m_min_x=m_range[0]-'a';

    if(m_key[0]>='A' && m_key[0]<='Z')
      m_symbol=m_key[0]-'A';
    else
      m_symbol=m_key[0]-'a';

    //Only single column allowed
    if(m_range[4]>='A' && m_range[4]<='Z')
      m_max_x=m_range[4]-'A';
    else
      m_max_x=m_range[4]-'a';
   
    m_min_y=(m_range[1]-'0')*10+(m_range[2]-'0')-1;
    m_max_y=(m_range[5]-'0')*10+(m_range[6]-'0')-1;

    m_val_max=m_val_min=0;
    if( m_min_x>=0&&m_min_x<27 &&
	m_min_y>=0&&m_min_y<100&&m_max_y>=0&&m_max_y<100 && m_min_y<=m_max_y){

      for(i=m_min_y; i<=m_max_y; i++){
	if(m_book[m_min_x][i]){
	  if(m_book[m_min_x][i]->value>m_val_max)
	    m_val_max=m_book[m_min_x][i]->value;
	  if(m_book[m_min_x][i]->value<m_val_min)
	    m_val_min=m_book[m_min_x][i]->value;
	}
      }

      m_scale=135/(m_val_max-m_val_min)/1.2;

      //SelectObject(hdc, (HPEN)GetStockObject(WHITE_PEN));
      SelectObject (hdc, GetSystemFontObject());
      GetSystemPaletteEntries(hdc, 0, 16, (LPPALETTEENTRY)&pals);
      hpen = CreatePen(PS_SOLID, 1, RGB(0,0,0));
      oldpen = SelectObject(hdc, hpen);
      hbrush = CreateSolidBrush(RGB(255,255,255));
      oldbrush = SelectObject(hdc, hbrush);
      //SelectObject (hdc, (HPEN)GetStockObject(BLACK_PEN));
      color_old=SetBkColor(hdc, GetSysColor(COLOR_3DLIGHT));
      SetBkMode(hdc, TRANSPARENT);
 
      //draw x, y axises line
      MoveToEx(hdc, 10, 17, NULL);
      LineTo(hdc, 10, 145);
      MoveToEx(hdc, 10, 140-(0-m_val_min*1.2)*m_scale, NULL);
      LineTo(hdc, 160, 140-(0-m_val_min*1.2)*m_scale);

      //draw the data points
      MoveToEx(hdc, 25, 
	       140-(m_book[m_min_x][m_min_y]->value-m_val_min*1.2)*m_scale,
	       NULL);
      for(i=m_min_y; i<=m_max_y&&i<=m_min_y+8; i++){
	if(m_book[m_min_x][i]){
	  if(i%2==0){
	    hbrush=CreateSolidBrush(RGB(pals[i%16].peRed,
					pals[i%16].peGreen,pals[i%16].peBlue));
	  }else{
	    hbrush=CreateSolidBrush(RGB(pals[(i+8)%16].peRed,
					pals[(i+8)%16].peGreen,
					pals[(i+8)%16].peBlue));
	  }
	  SelectObject(hdc, hbrush);

	  //data points
	  Ellipse(hdc, 25+(i-m_min_y)*15-5,
		  140-(m_book[m_min_x][i]->value
		       -m_val_min*1.2)*m_scale-5,
		  25+(i-m_min_y)*15+5,
		  140-(m_book[m_min_x][i]->value
		       -m_val_min*1.2)*m_scale+5);
	  LineTo(hdc, 25+(i-m_min_y)*15,
		 140-(m_book[m_min_x][i]->value
		      -m_val_min*1.2)*m_scale);

	  Rectangle(hdc, 25+(i-m_min_y)*15, 140-(0-m_val_min*1.2)*m_scale-2,
		    25+(i-m_min_y)*15, 140-(0-m_val_min*1.2)*m_scale+3);
	  sprintf(buf,"%-6.6g",m_book[m_min_x][i]->value);
	  if((i-m_min_y)%2==0){
	    TextOut (hdc, 25+(i-m_min_y)*15-6,
		     145-(m_book[m_min_x][i]->value-m_val_min*1.2)*m_scale,
		     buf, -1);
	  }else{
	    TextOut (hdc, 25+(i-m_min_y)*15-6,
		     127-(m_book[m_min_x][i]->value-m_val_min*1.2)*m_scale,
		     buf, -1);
	  }
	}
      }

      TextOut(hdc, 0, 140-(0-m_val_min*1.2)*m_scale-6, "0", -1);

      //SelectObject (hdc, (HPEN)GetStockObject(BLACK_PEN));
      SetBkMode(hdc, TRANSPARENT);
      TextOut(hdc, 2, 148, "Symbol:", -1);
      MoveToEx(hdc,0,160,NULL);
      LineTo(hdc,160,160);
      MoveToEx(hdc,79,160,NULL);
      LineTo(hdc,79,240);

      //show the symbol keys...
      j=0;
      for(i=m_min_y; i<=m_max_y&& j<=m_min_y+8; i++){
	if(m_book[m_min_x][i]){
	  if(i%2==0){
	    hbrush=CreateSolidBrush(RGB(pals[i%16].peRed,
					pals[i%16].peGreen,pals[i%16].peBlue));
	  }else{
	    hbrush=CreateSolidBrush(RGB(pals[(i+8)%16].peRed,
					pals[(i+8)%16].peGreen,
					pals[(i+8)%16].peBlue));
	  }
	  SelectObject(hdc, hbrush);
	  Ellipse(hdc,1+(int)(j/5)*80,165+j%5*15,
		    ((int)j/5)*80+10,173+j%5*15);
	  sprintf(buf, "%3.3g",m_book[m_min_x][i]->value);
	  if(m_book[m_symbol][i]){
	    sprintf(buf, "%3.3g %-6.6s",
		    m_book[m_min_x][i]->value,
		    m_book[m_symbol][i]->disp);
	  }
	  SelectObject (hdc, GetSystemFontObject());
	  TextOut(hdc, 10+(int)(j/5)*80, 165+j%5*15, buf, -1);
	  j++;
	}
      }

      SelectObject(hdc, oldpen);
      SelectObject(hdc, oldbrush);
      DeleteObject(hpen);
      DeleteObject(hbrush);

      SetBkColor(hdc, color_old);
    }else{
      //MessageBox(NULL,"Input range error...", "Error", MB_OK | MB_ICONERROR);
      EndPaint (hDlg, &ps);
      EndDialog (hDlg, IDOK);
      return TRUE;
    }
    
    //ReleaseDC(hDlg,hdc);
    EndPaint (hDlg, &ps);

    //InvalidateRect(GetDlgItem (hDlg, IDC_FILE), NULL, TRUE);
    //InvalidateRect(GetDlgItem (hDlg, IDC_LIST), NULL, TRUE);
    //SetFocus(GetDlgItem (hDlg, IDC_RANGE));
    break;

  case WM_COMMAND:
    switch ( LOWORD( wParam ) ){      
    case IDC_OK:
      EndDialog (hDlg, IDOK);
      return TRUE;
      break;    
    }
  default:
    return DefWindowProc( hDlg, message, wParam, lParam );
  }
  
  return 0;
}

static LRESULT CALLBACK BarProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){

  HDC hdc;
  PAINTSTRUCT ps;
  HBRUSH hbrush, oldbrush;
  HPEN hpen, oldpen;
  COLORREF color_old;

  int m_min_x, m_min_y;
  int m_max_x, m_max_y;
  int m_symbol;

  int i, j;
  //double m_axis_max, m_axis_min; // max & min value of Y-axis
  double m_val_max, m_val_min, m_scale;

  char buf[128];

  switch ( message ){
  case WM_INITDIALOG:
    SetWindowText(hDlg, "Line");
#if GVPDA
    SetScreenMode(4);      // 4bpp
    ResetScreen();
#else
    for(i=0; i<16; i++){
      pals[i].peRed=16*i;
      pals[i].peGreen=16*i;
      pals[i].peBlue=16*i;
    }
#endif
    break;

  case WM_PAINT:
    hdc = BeginPaint (hDlg, &ps);
    
    if(m_range[0]>='A' && m_range[0]<='Z')
      m_min_x=m_range[0]-'A';
    else
      m_min_x=m_range[0]-'a';

    if(m_key[0]>='A' && m_key[0]<='Z')
      m_symbol=m_key[0]-'A';
    else
      m_symbol=m_key[0]-'a';

    //Only single column allowed
    if(m_range[4]>='A' && m_range[4]<='Z')
      m_max_x=m_range[4]-'A';
    else
      m_max_x=m_range[4]-'a';
   
    m_min_y=(m_range[1]-'0')*10+(m_range[2]-'0')-1;
    m_max_y=(m_range[5]-'0')*10+(m_range[6]-'0')-1;

    m_val_max=m_val_min=0;
    if( m_min_x>=0&&m_min_x<27 &&
	m_min_y>=0&&m_min_y<100&&m_max_y>=0&&m_max_y<100 && m_min_y<=m_max_y){

      for(i=m_min_y; i<=m_max_y; i++){
	if(m_book[m_min_x][i]){
	  if(m_book[m_min_x][i]->value>m_val_max)
	    m_val_max=m_book[m_min_x][i]->value;
	  if(m_book[m_min_x][i]->value<m_val_min)
	    m_val_min=m_book[m_min_x][i]->value;
	}
      }

      m_scale=135/(m_val_max-m_val_min)/1.2;

      //SelectObject(hdc, (HPEN)GetStockObject(WHITE_PEN));
      SelectObject (hdc, GetSystemFontObject());
      GetSystemPaletteEntries(hdc, 0, 16, (LPPALETTEENTRY)&pals);
      hpen = CreatePen(PS_SOLID, 1, RGB(0,0,0));
      oldpen = SelectObject(hdc, hpen);
      hbrush = CreateSolidBrush(RGB(255,255,255));
      oldbrush = SelectObject(hdc, hbrush);
      //SelectObject (hdc, (HPEN)GetStockObject(BLACK_PEN));
      color_old=SetBkColor(hdc, GetSysColor(COLOR_3DLIGHT));
      SetBkMode(hdc, TRANSPARENT);
 
      //draw x, y axises line
      MoveToEx(hdc, 10, 17, NULL);
      LineTo(hdc, 10, 145);
      MoveToEx(hdc, 10, 140-(0-m_val_min*1.2)*m_scale, NULL);
      LineTo(hdc, 160, 140-(0-m_val_min*1.2)*m_scale);

      //draw the data points
      for(i=m_min_y; i<=m_max_y&&i<=m_min_y+8; i++){
	if(m_book[m_min_x][i]){
	  if(i%2==0){
	    hbrush=CreateSolidBrush(RGB(pals[i%16].peRed,
					pals[i%16].peGreen,pals[i%16].peBlue));
	  }else{
	    hbrush=CreateSolidBrush(RGB(pals[(i+8)%16].peRed,
					pals[(i+8)%16].peGreen,
					pals[(i+8)%16].peBlue));
	  }
	  SelectObject(hdc, hbrush);

	  //data points
	  sprintf(buf,"%-6.6g",m_book[m_min_x][i]->value);
	  if(m_book[m_min_x][i]->value>=0){
	    Rectangle(hdc, 25+(i-m_min_y)*15-5,
		      140-(m_book[m_min_x][i]->value-m_val_min*1.2)*m_scale,
		      25+(i-m_min_y)*15+6, 140-(0-m_val_min*1.2)*m_scale+1);
	    TextOut(hdc, 25+(i-m_min_y)*15-6,
		     129-(m_book[m_min_x][i]->value-m_val_min*1.2)*m_scale,
		     buf, -1);
	  }else{
	    Rectangle(hdc, 25+(i-m_min_y)*15-5,140-(0-m_val_min*1.2)*m_scale,
		      25+(i-m_min_y)*15+6,
		      140-(m_book[m_min_x][i]->value-m_val_min*1.2)*m_scale);
	    TextOut(hdc, 25+(i-m_min_y)*15-6,
		     141-(m_book[m_min_x][i]->value-m_val_min*1.2)*m_scale,
		     buf, -1);
	  }

	  Rectangle(hdc, 25+(i-m_min_y)*15, 140-(0-m_val_min*1.2)*m_scale-2,
		    25+(i-m_min_y)*15, 140-(0-m_val_min*1.2)*m_scale+3);
	}
      }

      //origin point
      TextOut(hdc, 0, 140-(0-m_val_min*1.2)*m_scale-6, "0", -1);

      //SelectObject (hdc, (HPEN)GetStockObject(BLACK_PEN));
      SetBkMode(hdc, TRANSPARENT);
      TextOut(hdc, 2, 148, "Symbol:", -1);
      MoveToEx(hdc,0,160,NULL);
      LineTo(hdc,160,160);
      MoveToEx(hdc,79,160,NULL);
      LineTo(hdc,79,240);

      //show the symbol keys...
      j=0;
      for(i=m_min_y; i<=m_max_y&& j<=m_min_y+8; i++){
	if(m_book[m_min_x][i]){
	  if(i%2==0){
	    hbrush=CreateSolidBrush(RGB(pals[i%16].peRed,
					pals[i%16].peGreen,pals[i%16].peBlue));
	  }else{
	    hbrush=CreateSolidBrush(RGB(pals[(i+8)%16].peRed,
					pals[(i+8)%16].peGreen,
					pals[(i+8)%16].peBlue));
	  }
	  SelectObject(hdc, hbrush);
	  Rectangle(hdc,2+(int)(j/5)*80,167+j%5*15,
		    ((int)j/5)*80+9,173+j%5*15);
	  sprintf(buf, "%3.3g",m_book[m_min_x][i]->value);
	  if(m_book[m_symbol][i]){
	    sprintf(buf, "%3.3g %-6.6s",
		    m_book[m_min_x][i]->value,
		    m_book[m_symbol][i]->disp);
	  }
	  SelectObject (hdc, GetSystemFontObject());
	  TextOut(hdc, 10+(int)(j/5)*80, 165+j%5*15, buf, -1);
	  j++;
	}
      }

      SelectObject(hdc, oldpen);
      SelectObject(hdc, oldbrush);
      DeleteObject(hpen);
      DeleteObject(hbrush);

      SetBkColor(hdc, color_old);
    }else{
      //MessageBox(NULL,"Input range error...", "Error", MB_OK | MB_ICONERROR);
      EndPaint (hDlg, &ps);
      EndDialog (hDlg, IDOK);
      return TRUE;
    }
    
    //ReleaseDC(hDlg,hdc);
    EndPaint (hDlg, &ps);

    //InvalidateRect(GetDlgItem (hDlg, IDC_FILE), NULL, TRUE);
    //InvalidateRect(GetDlgItem (hDlg, IDC_LIST), NULL, TRUE);
    //SetFocus(GetDlgItem (hDlg, IDC_RANGE));
    break;

  case WM_COMMAND:
    switch ( LOWORD( wParam ) ){      
    case IDC_OK:
      EndDialog (hDlg, IDOK);
      return TRUE;
      break;    
    }
  default:
    return DefWindowProc( hDlg, message, wParam, lParam );
  }
  
  return 0;
}

static LRESULT CALLBACK JumpProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){
  static HWND hwndOld;

  HWND	hedit;
  int nlen=0;
  int m_min_x, m_min_y;
  char m_id[8]="";

  switch ( message ){
  case WM_INITDIALOG:
    selectXime (XIME_ASCII);
    hwndOld = (HWND)setXime(hDlg);
    SendMessage(GetDlgItem (hDlg, IDC_RANGE), EM_LIMITTEXT, 10, 0);
    SetFocus(GetDlgItem (hDlg, IDC_RANGE));

    break;

  case WM_PAINT:
    //InvalidateRect(GetDlgItem (hDlg, IDC_FILE), NULL, TRUE);
    //InvalidateRect(GetDlgItem (hDlg, IDC_LIST), NULL, TRUE);
    SetFocus(GetDlgItem (hDlg, IDC_RANGE));
    break;

  case WM_COMMAND:
    switch ( LOWORD( wParam ) ){      
    case IDC_OK:
      // Get cell ID
      hedit = GetDlgItem(hDlg, IDC_RANGE);
      nlen = GetWindowText(hedit, m_id, 7);
      if (nlen <= 0) {
	SetFocus (hedit);
	return TRUE;
      }

      if(m_id[0]>='A' && m_id[0]<='Z')
	m_min_x=m_id[0]-'A';
      else
	m_min_x=m_id[0]-'a';
   
      m_min_y=(m_id[1]-'0')*10+(m_id[2]-'0')-1;

      if(m_min_x>=0&&m_min_x<27 && m_min_y>=0&&m_min_y<100){

	m_col=m_min_x;
	m_row=m_min_y;
	
	if(m_col+n_cols>26){
	  m_col-=(m_col+n_cols)-26;
	}

	if(m_row+n_rows>100){
	  m_row-=(m_row+n_rows)-100;
	}

	setXime(hwndOld);

	n_rows=13;
	selectXime(XIME_NONE);
	
	EndDialog (hDlg, IDOK);
	SetDlgItemText(hwndOld, ID_EDIT, "");
 
	return TRUE;
      }else{
	SetFocus(GetDlgItem(hDlg, IDC_RANGE));
	return TRUE;
      }
      return TRUE;
      
    case IDC_CANCEL:
      setXime(hwndOld);
      //ShowWindow(hXime, SW_HIDE);
      SetDlgItemText(hwndOld, ID_EDIT, "");
      EndDialog(hDlg, IDCANCEL);
      return IDCANCEL;
    }
    return 0;
  }
  
  return DefWindowProc( hDlg, message, wParam, lParam );
}
